use std::{
    cell::RefCell,
    collections::{HashMap, HashSet},
};

use self::{
    command::{Command, InternalCommand},
    r#trait::{InternalTrait, Trait},
    r#type::{InternalType, Type},
    typedef::TypeDef,
};

pub mod command;
pub mod r#trait;
pub mod r#type;
pub mod typedef;

pub struct GlobalNamespace {
    types: RefCell<Vec<InternalType>>,
    traits: RefCell<Vec<InternalTrait>>,
    type_namespace: RefCell<HashMap<String, TypeNamespaceId>>,
    commands: RefCell<Vec<InternalCommand>>,
    data_namespace: RefCell<HashMap<String, DataNamespaceId>>,
}

pub type TypeId = usize;
pub type TraitId = usize;
pub type CommandId = usize;

enum TypeNamespaceId {
    Types(TypeId),
    Traits(TraitId),
}

enum DataNamespaceId {
    Commands(CommandId),
}

impl GlobalNamespace {
    pub fn init() -> Self {
        Self {
            types: RefCell::new(Vec::new()),
            traits: RefCell::new(Vec::new()),
            type_namespace: RefCell::new(HashMap::new()),
            commands: RefCell::new(Vec::new()),
            data_namespace: RefCell::new(HashMap::new()),
        }
    }

    /// register a type to the namespace
    ///
    /// # Errors
    ///
    /// Will return `NsRegistrationError::NameAlreadyExists` if the desired name is already in use
    pub fn register_type(&self, name: &str) -> Result<Type, NsRegistrationError> {
        if self.type_namespace.borrow().contains_key(name) {
            Err(NsRegistrationError::NameAlreadyExists)
        } else {
            self.types.borrow_mut().push(InternalType {
                traits: RefCell::new(HashSet::new()),
                name: name.to_owned(),
            });
            let id = self.types.borrow().len() - 1;
            let _ = self
                .type_namespace
                .borrow_mut()
                .insert(name.to_owned(), TypeNamespaceId::Types(id));
            Ok(Type {
                id,
                namespace: self,
            })
        }
    }

    /// register a trait to the namespace
    ///
    /// # Errors
    ///
    /// Will return `NsRegistrationError::NameAlreadyExists` if the desired name is already in use
    pub fn register_trait(&self, name: &str) -> Result<Trait, NsRegistrationError> {
        if self.type_namespace.borrow().contains_key(name) {
            Err(NsRegistrationError::NameAlreadyExists)
        } else {
            self.traits.borrow_mut().push(InternalTrait {
                types: RefCell::new(HashSet::new()),
                name: name.to_owned(),
            });
            let id = self.traits.borrow().len() - 1;
            let _ = self
                .type_namespace
                .borrow_mut()
                .insert(name.to_owned(), TypeNamespaceId::Traits(id));
            Ok(Trait {
                id,
                namespace: self,
            })
        }
    }

    /// register a command to the namespace
    ///
    /// # Errors
    ///
    /// Will return `NsRegistrationError::NameAlreadyExists` if the desired name is already in use
    pub fn register_command<'a>(
        &self,
        name: &'a str,
        // TODO: refacto with builder pattern maybe?
        generics: Vec<(&'a str, Vec<&Trait>)>,
        input: Option<&'a TypeDef>,
        output: Option<&'a TypeDef>,
    ) -> Result<Command, NsRegistrationError<'a>> {
        if self.data_namespace.borrow().contains_key(name) {
            Err(NsRegistrationError::NameAlreadyExists)
        } else {
            let mut internal_generics_namespace: HashMap<String, Vec<TraitId>> = HashMap::new();

            for (name, conditions) in generics {
                if internal_generics_namespace.contains_key(name) {
                    return Err(NsRegistrationError::GenericNameAlreadyExists);
                }
                internal_generics_namespace
                    .insert(name.to_owned(), conditions.iter().map(|t| t.id).collect());
            }

            if let Some(def) = input {
                def.check_generics_exist(&internal_generics_namespace)
                    .map_err(NsRegistrationError::UnregisteredGenericsUsed)?;
            }

            if let Some(def) = output {
                def.check_generics_exist(&internal_generics_namespace)
                    .map_err(NsRegistrationError::UnregisteredGenericsUsed)?;
            }

            self.commands.borrow_mut().push(InternalCommand {
                name: name.to_owned(),
                input: input.map(std::convert::Into::into),
                output: output.map(std::convert::Into::into),
            });
            let id = self.traits.borrow().len() - 1;
            let _ = self
                .data_namespace
                .borrow_mut()
                .insert(name.to_owned(), DataNamespaceId::Commands(id));
            Ok(Command {
                id,
                namespace: self,
            })
        }
    }

    pub fn get_type(&self, id: TypeId) -> Option<Type> {
        (self.types.borrow().len() > id).then_some(Type {
            id,
            namespace: self,
        })
    }

    pub fn get_trait(&self, id: TraitId) -> Option<Trait> {
        (self.traits.borrow().len() > id).then_some(Trait {
            id,
            namespace: self,
        })
    }

    pub fn get_command(&self, id: CommandId) -> Option<Command> {
        (self.commands.borrow().len() > id).then_some(Command {
            id,
            namespace: self,
        })
    }

    pub fn get_type_by_name(&self, name: &str) -> Option<Type> {
        if let Some(TypeNamespaceId::Types(id)) = self.type_namespace.borrow().get(name) {
            Some(Type {
                id: *id,
                namespace: self,
            })
        } else {
            None
        }
    }

    pub fn get_trait_by_name(&self, name: &str) -> Option<Trait> {
        if let Some(TypeNamespaceId::Traits(id)) = self.type_namespace.borrow().get(name) {
            Some(Trait {
                id: *id,
                namespace: self,
            })
        } else {
            None
        }
    }

    pub fn get_command_by_name(&self, name: &str) -> Option<Command> {
        if let Some(DataNamespaceId::Commands(id)) = self.data_namespace.borrow().get(name) {
            Some(Command {
                id: *id,
                namespace: self,
            })
        } else {
            None
        }
    }
}

#[derive(Debug)]
pub enum NsRegistrationError<'a> {
    NameAlreadyExists,
    GenericNameAlreadyExists,
    UnregisteredGenericsUsed(Vec<&'a str>),
}
