use std::collections::HashMap;
use std::fmt::Debug;
use std::fmt::Display;

use super::TraitId;
use super::TypeId;

use super::GlobalNamespace;

use super::r#type::Type;

pub enum TypeDef<'a> {
    Type(Type<'a>),
    Generic(String),
    List(Vec<TypeDef<'a>>),
    Record(Vec<(String, TypeDef<'a>)>),
}

impl<'a> TypeDef<'a> {
    pub(super) fn check_generics_exist(
        &'a self,
        map: &HashMap<String, Vec<TraitId>>,
    ) -> Result<(), Vec<&'a str>> {
        match self {
            TypeDef::Type(_) => Ok(()),
            TypeDef::Generic(name) => {
                if map.contains_key(name) {
                    Ok(())
                } else {
                    Err(vec![name])
                }
            }
            TypeDef::List(defs) => {
                let r = defs
                    .iter()
                    .map(|def| def.check_generics_exist(map))
                    .filter_map(|check_res| {
                        if let Err(invalid_names) = check_res {
                            Some(invalid_names)
                        } else {
                            None
                        }
                    })
                    .reduce(|mut acc, mut errs| {
                        acc.append(&mut errs);
                        acc
                    });

                match r {
                    Some(e) => Err(e),
                    None => Ok(()),
                }
            }
            TypeDef::Record(rec) => {
                let r = rec
                    .iter()
                    .map(|(_n, def)| def.check_generics_exist(map))
                    .filter_map(|check_res| {
                        if let Err(invalid_names) = check_res {
                            Some(invalid_names)
                        } else {
                            None
                        }
                    })
                    .reduce(|mut acc, mut errs| {
                        acc.append(&mut errs);
                        acc
                    });

                match r {
                    Some(e) => Err(e),
                    None => Ok(()),
                }
            }
        }
    }
    pub(super) fn from_internal(ns: &'a GlobalNamespace, def: &InternalTypeDef) -> TypeDef<'a> {
        match def {
            InternalTypeDef::Single(id) => TypeDef::Type(
                ns.get_type(*id)
                    .expect("Unregistered internal type ID. This is a bug."),
            ),
            InternalTypeDef::Generic(name) => TypeDef::Generic(name.clone()),
            InternalTypeDef::List(list) => TypeDef::List(
                list.iter()
                    .map(|def| Self::from_internal(ns, def))
                    .collect(),
            ),
            InternalTypeDef::Record(rec) => TypeDef::Record(
                rec.iter()
                    .map(|(name, def)| (name.clone(), Self::from_internal(ns, def)))
                    .collect(),
            ),
        }
    }
}

impl Display for TypeDef<'_> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            TypeDef::Type(t) => Display::fmt(&t, f),
            TypeDef::Generic(name) => Display::fmt(name, f),
            TypeDef::List(l) => {
                f.write_str("[ ")?;
                if let Some(first) = l.first() {
                    Display::fmt(&first, f)?;
                }
                for item in l.iter().skip(1) {
                    f.write_str(", ")?;
                    Display::fmt(&item, f)?;
                }
                f.write_str(" ]")
            }
            TypeDef::Record(rec) => {
                f.write_str("{ ")?;
                for (i, item) in rec.iter().enumerate() {
                    if i != 0 {
                        f.write_str(", ")?;
                    }
                    f.write_fmt(format_args!("{}: {}", item.0, item.1))?;
                }
                f.write_str(" }")
            }
        }
    }
}

impl Debug for TypeDef<'_> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            TypeDef::Type(t) => Debug::fmt(&t, f),
            TypeDef::Generic(name) => Debug::fmt(name, f),
            TypeDef::List(l) => {
                f.write_str("[ ")?;
                for (i, item) in l.iter().enumerate() {
                    if i != 0 {
                        f.write_str(", ")?;
                    }
                    Debug::fmt(&item, f)?;
                }
                f.write_str(" ]")
            }
            TypeDef::Record(rec) => {
                f.write_str("{ ")?;
                for (i, item) in rec.iter().enumerate() {
                    if i != 0 {
                        f.write_str(", ")?;
                    }
                    f.write_fmt(format_args!("{:?}: {:?}", item.0, item.1))?;
                }
                f.write_str(" }")
            }
        }
    }
}

impl<'a> From<Type<'a>> for TypeDef<'a> {
    fn from(value: Type<'a>) -> Self {
        TypeDef::Type(value)
    }
}

pub(super) enum InternalTypeDef {
    Single(TypeId),
    Generic(String),
    List(Vec<InternalTypeDef>),
    Record(Vec<(String, InternalTypeDef)>),
}

impl From<&TypeDef<'_>> for InternalTypeDef {
    fn from(value: &TypeDef) -> Self {
        match value {
            TypeDef::Type(val) => Self::Single(val.id),
            // TODO: rewrite this to be better
            TypeDef::Generic(name) => Self::Generic(name.to_owned()),
            TypeDef::List(list) => Self::List(list.iter().map(std::convert::Into::into).collect()),
            TypeDef::Record(rec) => Self::Record(
                rec.iter()
                    .map(|(name, typ)| (name.to_owned(), typ.into()))
                    .collect(),
            ),
        }
    }
}
